#!/usr/bin/env perl
# vim:aw:
# $Id: nsc.pl,v 1.20 2004/03/08 16:32:31 gothix Exp $
#
# nsc.pl - curses-based console monitor for Nagios
# 000206shj 
#
# Copyright (C) Stig H. Jacobsen & Gothix 2000-2004
#

# --- Bugs ----------------------------------------------------------------
#  - none listed


# --- Stuff ---------------------------------------------------------------

use strict;
use warnings;

use Carp;
use Curses;

eval "use Data::Dumper;";     #debug

eval "use gothix::cpusage;";
my $cpusage = !($@);          #true if mod is available


# --- Options and their default values ------------------------------------

my $VERSION = "v0.71";
my $myname = 'nsc';
my $me = $myname;

my $fnConfig = "$ENV{'HOME'}/.nsc.conf";


# =========================================================================
#                           G L O B A L S         
# =========================================================================


# --- Global variables ----------------------------------------------------

my $MIN_LINES = 24;
my $MIN_COLS = 80;

my $ns_app = '???';      #ns=netsaint, nag=nagios
my $ns_app_long = '?unknown?';   #long version of ns_app from status.log

my $opt_colour;
my $num_msglines;
my $first_msgline;
my $BOTLINE;

#current colour pair
my $pairno;

my $lastBeep = time();        #now on startup, to avoid initial bell
my $listoffs = 0;
my $display_dirty = 1;
my $num_displays = 0;         #number of times service list has been shown
my %AT = ();
my $myself;

my %sVals = (
      'CRITICAL',    0,
      'HOST DOWN',   1,
      'UNREACHABLE', 2,
      'WARNING',     3,
      'UNKNOWN',     4,
      'PENDING',     5,
      'RECOVERY',    6,
      'OK',          7
      );

$ENV{'VISUAL'} = $ENV{'EDITOR'} if !defined($ENV{'VISUAL'});
$ENV{'VISUAL'} = 'vi' if !defined($ENV{'VISUAL'});
$ENV{'PAGER'} = 'more' if !defined($ENV{'PAGER'});


# --- Global status data --------------------------------------------------

my %hState;    #hosts states
my %sState;    #service states

my @statusList;               #list of hash-refs
my $tm_status_log = 0;     #last load time from disk

my $pgmUp;
my $pgmState;


# label: text used for headings
# width: factory standard with for 80 cols
# cw: current width adjusted for current window width
# right: true/set if content is to be right-justified (numbers)
my %DataCols = (
   '0_node'      => {
      'label'  => 'Node',
      'width'  => 8
      },
   '1_service'   => {
      'label'  => 'Service',
      'width'  => 14
      },
   '2_status'    => {
      'label'  => ' Status',
      'width'  => 8
      },
   '3_updated'   => {
      'label'  => 'Updat',
      'width'  => 3
      },
   '4_tries'     => {
      'label'  => 'Try',
      'width'  => 3,
      'right'  => 1
      },
   '5_message'   => {
      'label'  => 'Service information',
      'width'  => -1
      }
);

#init cw col
foreach my $k (keys %DataCols) {
   $DataCols{$k}->{'cw'} = $DataCols{$k}->{'width'};
}


# --- Debugging support ---------------------------------------------------

{ #localize debug stack and friends

my @DebugStack = ();
my @DebugTM = ();
my $DebugStackSize = 500;

sub show_debug {

   return unless cfg('debug');

   drawScreen();

   my $line = 0;
   for (my $i=$#DebugStack; $i>=0 && $line<=$BOTLINE; $line++, $i--) {
      mLine($line, AT('normal'), tdif(time(), $DebugTM[$i]) . ' ' . $DebugStack[$i]);
      }

   botline("Press any key to continue..");
   rdKey(60);

} #show_debug


sub dm {

   return unless cfg('debug');

   if ($#DebugStack > $DebugStackSize) {
      #recycle to avoid consuming all swap within the next 6 years
      @DebugStack = $DebugStack[0..int($DebugStackSize/2)];
      dm('Debug stack reset');
   }
   push(@DebugStack, join('', @_));
   push(@DebugTM, time());

} #dm


} #end localization


#Debug output to stderr
sub d {

   return unless cfg('debug');

   print STDERR @_, "\n";

} #d


# =========================================================================
#                           C U R S E S           
# =========================================================================

# --- Curses replacement functions ----------------------------------------
#
#
sub mvaddch {
   my($y,$x,$ch) = @_;
   move($y,$x);
   addch($ch);
}

sub mvaddstr {
   my($y,$x,$str) = @_;
   move($y,$x);
   addstr($str);
}


# --- Low level "graphics" ------------------------------------------------

sub AT {
   die "$_[0] is not a defined colour" if !defined($AT{$_[0]});
   return $AT{$_[0]};
}


sub gotoxy {
   my($x,$y) = @_;

   croak "gotoxy($x,$y): X out of range!" unless (($x >= 0) && ($x <= $COLS));
   croak "gotoxy($x,$y): Y out of range!" unless (($y >= 0) && ($y <= $LINES));

   move($y,$x);
}


sub rdKey {

   #set attr, otherwise cursor will paint in last(?) colour
   #TODO: it doesn't normalize :-(
   attron(AT('normal'));

   gotoxy($COLS-1,$BOTLINE);
   refresh;
   attroff(AT('normal'));

   timeout(1000*$_[0]) if defined($_[0]);
   my $res = getch();

   if (defined($res) && ($res eq '-1')) {
      return undef;
   } else {
      return $res;
   }

} #rdKey


# at(int x, int y, const char *str)
sub at {

   my($x,$y,$str) = @_;

   if (substr($str,0,1) eq '!') {         #blank rest of line?
      $str = substr($str, 1);
      $str .= (' ' x ($COLS - length($str) - $x));
   } elsif (substr($str,0,1) eq '>') {    #relative to right side of display?
      $str = substr($str, 1);
      $x = $COLS - length($str) - $x;
      }

   #please control yourself or I will (we never wrap)
   if (length($str) > ($COLS-$x)) {
      dm("truncated '$str'");
      $str = substr($str, 0, ($COLS-$x));
   }

   mvaddstr($y,$x,$str);

   return length($str);

} #at


# aat(int x, int y, attrib, const char *str)
sub aat {
  attron($_[2]);
  at($_[0], $_[1], $_[3]);
  attroff($_[2]);
  return length($_[3]);
}


# --- UI stuff ------------------------------------------------------------

sub botline {
   aat(0, $BOTLINE, AT('botline'), sprintf('%-40.40s', $_[0]));
   refresh();
}


sub mLine {
   my ($line,$attrib,$msg) = @_;
   $attrib = AT('normal') if (!defined($attrib));
   aat(0, $line+$first_msgline, $attrib, "!$msg");
} #mLine


# =========================================================================
#                   C O N F I G U R A T I O N
# =========================================================================


# --- Configuration stuff -------------------------------------------------


{ #brace to localize %CFG and friends


my %NSC_KEYWORDS = (
   'nslog',          '/usr/local/nagios/var/status.log',
   'hostcfg',        '/usr/local/nagios/etc/services.cfg',
   'reloadcmd',      '/etc/rc.d/init.d/nagios reload',
   'showall',        '1',
   'details',        '1',
   'reverse',        '0',
   'bell',           '1',
   'bell_interval',  '180',    #in seconds
   'debug',          '0',
   'updated',        0,             #last write, time(2)
   'cpumeter',       0,
   'roll',           30,             #how often to roll, -1=disable, else in seconds
   'roll_wait',      60,             #seconds of no user input before roll starts
   'version',        $VERSION,
   'usercmd_T',      'top',

   #monochrome attributes
   'mono.botline'    => 'bold',
   'mono.critical'   => 'reverse + bold',
   'mono.dim'        => 'dim',
   'mono.heading'    => 'underline',
   'mono.isok'       => '!normal',
   'mono.noisy'      => 'reverse + bold + blink',
   'mono.normal'     => 'normal',
   'mono.notok'      => 'bold',
   'mono.total'      => 'bold',
   'mono.warning'    => 'reverse',

   #colour attributes
   'colour.botline'  => '!normal',
   'colour.critical' => 'white on red',
   'colour.dim'      => '!normal',
   'colour.heading'  => 'white on blue + bold',
   'colour.isok'     => 'white on green',
   'colour.noisy'    => 'white on red',
   'colour.normal'   => 'white on blue',
   'colour.notok'    => '!warning',
   'colour.total'    => '!botline',
   'colour.warning'  => 'white on yellow + bold',

   #obselete keywords
   'upd_freq', '2',

);



my %CFG;                #configuration items
my $cfg_updated = 0;    #time_t of last %CFG change
my $cfg_loaded = 0;     #time_t of last %CFG load from disk
my $cfg_changes = 0;    #number of changes to in-memory %CFG


# cfg($varname) returns value of var
# cfg($varname, $newval) sets value of var, returns new value
sub cfg {

   my($var,$newval) = @_;

   if (defined($newval)) {
      if (defined($CFG{$var})
          && ($CFG{$var} ne $newval)) {   #update only on changes..
         $CFG{$var} = $newval;
         $cfg_updated = time();
         $cfg_changes++;
      } else {
         print STDERR "no cfg change: $var => $newval\n" if $CFG{'debug'};
      }
   }
   return $CFG{$var};

} #cfg


# --- Colour configuration ------------------------------------------------

{ #localize %attrColours, etc.

my %attrMono = (
   'bold'      => A_BOLD,
   'blink'     => A_BLINK,
   'normal'    => A_NORMAL,
   'reverse'   => A_REVERSE,
   'dim'       => A_DIM,
   'underline' => A_UNDERLINE,
);

my %attrColour = (
   'white'     => COLOR_WHITE,
   'blue'      => COLOR_BLUE,
   'red'       => COLOR_RED,
   'yellow'    => COLOR_YELLOW,
   'green'     => COLOR_GREEN,
   'black'     => COLOR_BLACK,
   'magenta'   => COLOR_MAGENTA,
   'default'   => -1,               #ncurses specific
);


# Return Curses attribute value (for attron()) for named nsc highlighting type
#   $a = decode_mono('botline');    #returns perhaps A_REVERSE
#   attron($a);
sub decode_mono {

   my ($attr) = @_;

   my $ky = 'mono.' . $attr;

   croak "Unknown monochrome highlighting '$attr'"
      unless defined($CFG{$ky});

   #recurse for indirect colours
   if ($CFG{$ky} =~ /^!(.+)$/) {
      return decode_mono($1);
   }

   my $val = $CFG{$ky};
   $val =~ s/\s+//g;
   my @attrs = split(/\+/, $val);

   my $res = undef;
   foreach my $as (@attrs) {
      croak "Unknown monochrome attribute '$as'"
         unless defined($attrMono{$as});
      my $a = $attrMono{$as};
      #$a = $attrMono{$a} if defined($attrMono{$a});    #indirect, i.e. mono.heading = 'normal'
      if (defined($res)) {
         $res |= $a;
      } else {
         $res = $a;
      }
   }

   return $res;

} #decode_mono


# decode_colour()
# As decode_mono(), but for colour attributes
sub decode_colour {

   my ($attr) = @_;
   my $res = undef;

   my $ky = 'colour.' . $attr;

   croak "Unknown colour highlighting '$attr'"
      unless defined($CFG{$ky});

   #recurse for indirect colours
   if ($CFG{$ky} =~ /^!(.+)$/) {
      return decode_colour($1);
   }

   # $val = 'white on blue', optionally '+ bold'
   my $val = $CFG{$ky};

   my $cpair = ++$pairno;

   my ($fore,$back);

   if (!(($fore,$back) = split(/\s+on\s+/, $val))) {
      $fore = $val;
      $back = 'default';
   }

   my @plusses = ();
   my @temp;
   ($fore,@temp) = split(/[\s+]+/, $fore);
   push(@plusses, @temp);
   ($back,@temp) = split(/[\s+]+/, $back);
   push(@plusses, @temp);
   init_pair($cpair, $attrColour{$fore}, $attrColour{$back});
   $res = COLOR_PAIR($cpair);
   foreach my $plus (@plusses) {
      my $a = $attrMono{$plus} || croak "Unknown +attribute '$plus' ($attr)";
      $res |= $a;
   }

   return $res;

} #decode_colour


} #end localization


# --- Configuration -------------------------------------------------------

sub SaveConfig {

   open(CFG, ">$fnConfig") || die("Can't open $fnConfig for writing: $!");
   foreach my $ky (sort keys %NSC_KEYWORDS) {
      print CFG "$ky=$CFG{$ky}\n";
      }
   print CFG "updated=" . localtime() . sprintf(", %d changes\n", $cfg_changes);
   close(CFG);
   $cfg_changes = 0;
   $cfg_loaded = time();

} #SaveConfig


sub LoadConfig {

   my $errcnt = 0;

   %CFG = ();

   #TODO: Version-check on loaded configfile?
   if (open(CFG, "$fnConfig")) {
      while (<CFG>) {
         chomp;
         my ($kw,$val) = split(/=/);
         if (defined($NSC_KEYWORDS{$kw})) {
            $CFG{$kw} = $val;
         } else {
            print STDERR "$me: Unknown keyword $kw\n";
            $errcnt++;
            }
         } #eof CFG
      close(CFG);
      }

   #apply default values
   foreach my $kw (keys %NSC_KEYWORDS) {
      $CFG{$kw} = $NSC_KEYWORDS{$kw}
         if (!defined($CFG{$kw}));
      }

   #override, cause this version to written when exiting
   $CFG{'version'} = $VERSION;

   $cfg_loaded = time();

   if ($errcnt) {
      print STDERR "\n*** $errcnt errors found in configuration file $fnConfig\n";
      print STDERR "Press Enter to continue ..";
      my $s = <STDIN>;
   }

} #LoadConfig


# --- End configuration ---------------------------------------------------


} #end configuration brace


# =========================================================================
#                   M I S C E L L A N O U S       
# =========================================================================


#$str = center('mystr', 10);
#return centered string for field $wid chars wide, space-padded, left and right
sub center {

   my($str, $wid) = @_;

   my $len = length($str);
   my $spc = int(($wid - $len) / 2);
   return $str if ($spc <= 0);

   my $res = (' ' x $spc) . $str;
   $res .= ' ' x ($wid - length($res));

   return $res;

} #center



# -------------------------------------------------------------------------

sub numeric {
   return 0 if (!defined($_[0]) || ($_[0] eq ''));
   return ($_[0] =~ /^\d+$/);
}


# -------------------------------------------------------------------------

sub xcmd {
   screen_down();
   system $_[0];
   screen_init();
   drawScreen();
} #xcmd


# --- tdif() --------------------------------------------------------------
# This is some of my weirder code
# Return difference between two time_t's as a string
# (ex. '1h23m40s'..)

{ #brace to localize %sv

   my %sv = (
      'Y', 365*24*60*60,
      'M', 30*24*60*60,
      'W', 7*24*60*60,
      'd', 24*60*60,
      'h', 60*60,
      'm', 60);

   sub tdif {

      my $dif;

      return '*undef*' if (!$_[0] || !$_[1]);

   #   #oh well..
   #  $dif = ($_[0]>$_[1])?$_[0]-$_[1]:$_[1]-$_[0];                

   # same shit, different code:
      $dif = $_[0]-$_[1];
      $dif = -$dif if ($dif<0);

      return "now"
         if ($dif == 0);

      my $res = '';
      my $details = 0;
      foreach my $key ('Y','M','W','d','h','m') {
         if ($dif > $sv{$key}) {
            my $i = int($dif / $sv{$key});
            $dif %= $sv{$key};
            $res .= "${i}$key";
            $details++;
            }
         last if ($details >= 2);
         }
      $res .= "${dif}s"
         if (($dif > 0) && ($details < 2));
      if (defined($_[2])) {   #wants just first element?
         #retain only first component
         $res = $1 if ($res =~ /^(\d+[a-z])\d/);
         }
      return $res;

   } #tdif

}


# =========================================================================
#                   S T A T I C   S C R E E N     
# =========================================================================

# --- Help ----------------------------------------------------------------

sub help {

   my @HLP = (
      '  h     This help',
      '  Space Next page',
      '  b     Previous page',
      '  ^     Move to top of service list',
      '  a     Toggle between showing all services and troubled ones',
      '  d     Toggle service details on/off',
      '  q     Quit',
      '  r     Reverse sort order',
      '  g     Toggle bell on/off',
      '  ^L    Redraw screen',
      '  C     Toggle colour on/off',
      '  V     View status.log',
      '  E     Edit services.cfg and reload',
      '',
      'nsc was written by Stig H. Jacobsen, <nsc@gothix.dk>',
      '',
      'Check http://freshmeat.net/projects/nsc/ for new versions',
      '',
      'Mailing list and support at http://groups.yahoo.com/group/nsc-users/',
      );
   
   drawScreen();

   my $i;
   for ($i=0; defined($HLP[$i]) && ($i<$BOTLINE); $i++) {
      mLine($i, AT('normal'), "$HLP[$i]");
      }

   botline("Press any key to continue..");
   rdKey(60);
   botline("");

} #help


# -------------------------------------------------------------------------

sub drawScreen {

   my @attribs = ( 
                   'botline', 'notok', 'critical', 'noisy', 'dim', 
                   'heading', 'total', 'warning', 'isok',
                   'normal'
                 );

   if ($opt_colour) {
      start_color();
      #colorpair 0 resets to zero?
      $pairno = 1;

      foreach my $attr (@attribs) {
         $AT{$attr} = decode_colour($attr);
      }
   } else {
      foreach my $attr (@attribs) {
         $AT{$attr} = decode_mono($attr);
      }
   }

   #force update with background colour
   attron(AT('normal'));
#   bkgd(A_NORMAL | AT('normal') | ' ');
   erase();
   for (my $i=0; $i<$LINES; $i++) {
      aat(0, $i, AT('normal'), ' ' x $COLS);
      }
   refresh();

   aat(0, 0, AT('dim'), "$ns_app_long / $myname $VERSION");

} #drawScreen


# --- screen_reset() ------------------------------------------------------
# Reset various nsc vars based on Curses info

sub screen_reset {

   croak "Too few lines in window, $MIN_LINES required, we have $LINES"
      unless ($LINES >= $MIN_LINES);
   croak "Too few columns in window, $MIN_COLS required, we have $COLS"
      unless ($COLS >= $MIN_COLS);

   $opt_colour = has_colors();

   $num_msglines = $LINES - 5;
   $first_msgline = 2;
   $BOTLINE = $LINES-1;

} #screen_reset


# --- screen_init() -------------------------------------------------------

sub screen_init {

   initscr(); cbreak(); noecho();
   nonl();
#&stdscr=warns, $stdscr=errs
#   intrflush(&stdscr, $FALSE);
#   keypad(&stdscr, $TRUE);

#   #wrap in eval, as this is ncurses specific
#   eval "use_default_colors()";
   
   screen_reset();

} #screen_init


sub screen_down {

   erase();
   refresh();
   endwin();

} #screen_down


# =========================================================================
#                   S E R V I C E   L I S T       
# =========================================================================


# --- Service list ordering -----------------------------------------------

sub by_prior {

   my $i = $sVals{$a->{'status'}} - $sVals{$b->{'status'}};
   if ($i == 0) {
      if (($i = ($a->{'node'} cmp $b->{'node'})) == 0) {
         $i = (lc($a->{'service'}) cmp lc($b->{'service'}));
         }
      }
   $i = -$i if (cfg('reverse'));
   return $i;

} #by_prior

sub sort_state {
   my @state = @_;
   @state = sort by_prior @state;
   return @state;
}


# --- Display single service line ------------------------------------

my ($lastNode,$lastState) = ('','');


sub dispService {

   my ($line,$row) = @_;

#   dm('type='.$row->{'type'});
#   foreach my $ky (sort keys %{$row}) {
#      dm("$ky => '$row->{$ky}'");
#   }
#   dm("-----");

   croak "dispService(): Row is undef" unless defined($row);

   return 0 if (!cfg('showall') && ($row->{'status'} eq 'OK'));

   my $date = '';
   my $change = $row->{'updated'};
   if (numeric($change)) {
      if ($change == 0) {
         $date = '';
      } else {
         $date = tdif(time, $change, 'y');
      }
   } else {
      dm("change='" . ($change?$change:'(undef)') . "'");
   }

   my $attrib;
   if ($row->{'status'} eq 'OK') {
      $attrib = AT('isok');
   } else {
      if ($row->{'status'} eq 'CRITICAL') {
         $attrib = AT('critical');
      } elsif ($row->{'status'} eq 'WARNING') {
         $attrib = AT('warning');
      } else {
         $attrib = AT('notok');
         }
      #TODO: use hState/sState instead of this...
      if ( ($row->{'status'} !~ /^(RECOVERY|OK|PENDING)$/) &&
           (cfg('bell') && ((time - $lastBeep) > cfg('bell_interval'))) ) {
         beep();
         $lastBeep = time;
         }
      }

   if (cfg('details')) {
      my $dispnode = $row->{'node'};
      $dispnode = ' .' if (($row->{'node'} eq $lastNode) && ($row->{'status'} eq $lastState));

      my $dispstate = $row->{'status'};
      $dispstate = '.' if (($row->{'node'} eq $lastNode) && ($row->{'status'} eq $lastState));
      $dispstate = center($dispstate, 8);

      $date = center($date, 3);
      my $attempts = center($row->{'attempts'}, 3);

      #show coloured status
      my $i = aat(0, $line+$first_msgline, AT('normal'),
                  sprintf("%-8.8s %-14.14s ", $dispnode,$row->{'service'} ));
      $i += aat($i, $line+$first_msgline, $attrib, sprintf("%-8.8s", $dispstate));
      $i += aat($i, $line+$first_msgline, AT('normal'),
                sprintf("! %3.3s %3.3s ", $date,$row->{'attempts'})) - 1;     #-1='!'

      my $left = $COLS - $i;

      my $serv_info = substr($row->{'info'}, 0, $left);
      $i += aat($i, $line+$first_msgline, AT('normal'), $serv_info);
   } else {
      mLine($line, $attrib,
            sprintf("%-14.14s %s", "$row->{'node'}:$row->{'service'}",$row->{'info'}));
      }

   ($lastNode,$lastState) = ($row->{'node'},$row->{'status'});

   return 1;

} #dispService


# --- Counts, sorts & weeds servicelist ------------------------------

sub getServiceList {
   my @res = ();

   foreach my $r (@statusList) {
      my $t = $r->{'type'} || croak "item has no type...\n" . Dumper($r);
      if ($t eq 'SERVICE') {
         push(@res, $r);
      } elsif ($t eq 'PROGRAM') {
         if ($ns_app eq 'ns') {
            # # NetSaint 0.0.7b6 Status File
            # [1002743117] PROGRAM;1002058577;2180;1;ACTIVE;1002058577;0;1002405600;1;1;1;0;0
            $pgmUp = tdif(time, $r->{'uptime'},'y');
            $pgmState = $r->{'status'};
         } elsif ($ns_app eq 'nag') {
            # # Nagios 1.0b6 Status File
            # [1041516783] PROGRAM;1037463573;27646;1;1041516768;1041116400;1;1;1;1;0;0;1;0
            #from nagions 1.06b6 src: snprintf(new_program_string,sizeof(program_string)-1,
            #      "[%lu] PROGRAM;%lu;%d;%d;%lu;%lu;%d;%d;%d;%d;%d;%d;%d;%d\n",
            #  0   (unsigned long)current_time,
            #  1   (unsigned long)_program_start,
            #  2   _nagios_pid,
            #  3   _daemon_mode,
            #  4   (unsigned long)_last_command_check,
            #  5   (unsigned long)_last_log_rotation,
            #  6   _enable_notifications,
            #  7   _execute_service_checks,
            #  8   _accept_passive_service_checks,
            #  9   _enable_event_handlers,
            # 10   _obsess_over_services,
            # 11   _enable_flap_detection,
            # 12   _enable_failure_prediction,
            # 13   _process_performance_data);
            $pgmUp = tdif(time, $r->{'uptime'},'y');
            #Nagios is running if status.log exists - when down, status.log
            #is removed and data saved to status.sav
            #(TODO: better support for this scheme)
            $pgmState = 'ACTIVE';
         } else {
            die "huh!? ($ns_app)";
            }
         }
      } #foreach @ol

   return sort_state(@res);

} #getServiceList


# =========================================================================
#                       S T A T U S   L O G           
# =========================================================================


# load_status_log($filename)
# Loads status.log unconditionally into @statusList.
sub load_status_log {

   my ($NSLOG) = @_;

   if (!open(STATUS, $NSLOG)) {
      botline("*** Can't open statuslog $NSLOG: $!");
      return undef;
      }

   #refill, please
   @statusList = ();
   while (<STATUS>) {
      next if (/^#/);
      chomp;
      my $row = {
         'raw' => $_
         };
      push(@statusList, $row);
   }

   close(STATUS);
   $tm_status_log = time();
   return 1;

} #load_status_log


# --------------------------------------------------------------------
# prep_status()
# Parses statusList->raw to appropriate places.

{

my $last_prep_tm = 0;
my @LongNames = ('node', 'service');
my %Longest = ();

sub prep_status {

   if ($last_prep_tm >= $tm_status_log) {
      return;  #no changes
   }

   foreach my $row (@statusList) {

      #dm('raw = ', $row->{'raw'});
      #show_debug();

      my @SVC = split(/;/, $row->{'raw'});

      $SVC[0] =~ /^\[(\d+)] (.*)$/ || croak "Format unrecognized: \"$SVC[0]\"";
      $row->{'tm'} = $1;
      $row->{'type'} = $2;

      if ($row->{'type'} eq 'SERVICE') {
         #[1078349554] SERVICE;dax;IMAP mail;CRITICAL;3/3;HARD;1078349453;1078349573;ACTIVE;1;1;1;1064879368;0;CRITICAL;0;0;0;11514636;1078338644;864;1;0;0;1;0;0.00;0;1;1;1;(Return code of 127 is out of bounds - plugin may be missing)
         ($row->{'node'},
         $row->{'service'},
         $row->{'status'},
         $row->{'attempts'},
         $row->{'updated'}) = (@SVC[1..4],$SVC[6]);
         $row->{'info'} = $SVC[$#SVC];
         $sState{$row->{'status'}}++;
      } elsif ($row->{'type'} eq 'HOST') {
         # [1078349554] HOST;localhost;UP;1078349454;1064880057;0;11513957;529;0;0;0;1;1;1;1;0;0.00;0;1;1;PING OK - Packet loss = 0%, RTA = 0.26 ms
         $row->{'node'} = $SVC[1];
         $row->{'status'} = $SVC[2];
         $row->{'info'} = $SVC[$#SVC];
         $hState{$row->{'status'}}++;
      } elsif ($row->{'type'} eq 'PROGRAM') {
         # [1078349554] PROGRAM;1076394004;2714;1;1078349539;1078009200;1;1;1;1;0;0;1;0
         ($row->{'uptime'}) = $SVC[1];
         $row->{'status'} = $SVC[3];
      } else {
         dm("unrecognized type '$row->{'type'}'");
         #otherwise ignore
         }

      # --- no content messing from here on ---
     
      if ($row->{'type'} eq 'SERVICE') {
         #Update column widths
         foreach my $lnm (@LongNames) {
            croak "$lnm?" unless defined($row->{$lnm});
            if (!defined($Longest{$lnm}) || (length($row->{$lnm}) > $Longest{$lnm})) {
               $Longest{$lnm} = length($row->{$lnm});
               }
            }
         }

   }

   dm('prep_status(): been there, done that');
   $last_prep_tm = time();

#   if ($COLS > $MIN_COLS) {
#      my $incr = int( ($COLS - $MIN_COLS) / ($#LongNames + 1 + 1) );
#      if ($incr > 0) {
#         $DataCols{"0_node"}->{cw} = $DataCols{"0_node"}->{width} + $incr;
#         $DataCols{"1_service"}->{cw} = $DataCols{"1_service"}->{width} + $incr;
#      }
#   }

} #prep_status

} #localize


# --------------------------------------------------------------------

sub dump_svclist {

   my (@list) = @_;

   foreach my $i (0..$#list) {
      print STDERR "-----\nitem $i:\n";
      print STDERR Dumper($list[$i]), "\n";
   }
   print STDERR "end-of-list\n";

} #dump_svclist


# --------------------------------------------------------------------

my $upd_pos = 0;

{ #localize data

my $last_service_tm = 0;

sub showServiceList {

   my (@svcList) = @_;

   if ($display_dirty) {
      dm('display dirty');
   } else {
      if ($tm_status_log <= $last_service_tm) {
         dm("noload, file not changed");
         return 0;
      }
      dm("load!");
   }

   # --- Show Nagios update/state
   my $s = '';
   if (defined($pgmState)) {
      if ($pgmState eq 'ACTIVE') {
         $s .= 'run ';
      } else {
         $s .= 'STOPPED '."($pgmState)";
         }
      }
   $s .= "$pgmUp "
      if (defined($pgmUp));
   $s =~ s/\s+$//g;
   {
      my $upd = tdif(time(), $tm_status_log, 'y');
      my $s2 = "$s $myself";
      aat(0, 1, ($pgmState eq 'ACTIVE')?AT('dim'):AT('noisy'), ">$s2");
      $upd_pos = length($s2);
   }

   # --- Show
   if (cfg('details')) {
      my $i = 0;
      my $out = '';
      foreach my $h (sort keys %DataCols) {
         my $hr = $DataCols{$h};
         last if ($i >= $COLS);
         my $fmt = '%' . (defined($hr->{right}) ? '' : '-');
         if ($hr->{cw} > 0) {
            $fmt .= sprintf("%d.%d", $hr->{cw}, $hr->{cw});
            }
         my $s = sprintf($fmt . 's ', $hr->{'label'});
         $i += length($s);
         $out .= $s;
      }
      if (1) {
         mLine(0, AT('heading'), $out);
      } else {
         mLine(0, AT('heading'),
               sprintf("%-8.8s %-14.14s %-8.8s %-3.3s %3.3s %s",
            'Node','Service','Status','Upd','Try','Service information'));
         }
   } else {
      mLine(0, AT('heading'),
            sprintf("%-14.14s %s", 
                    'Node:Service',
                    'Service information'));
      }

   ($lastNode,$lastState) = ('','');

   my $i = $listoffs;
   my $num_shown;

   if (!defined($svcList[$i])) {             #got shorter?(!)
      $i = $listoffs = 0;                    #reset as needed
      }

   for ($num_shown=0; ($num_shown<($num_msglines-1)) && defined($svcList[$i]); $i++) {
      if (dispService($num_shown + 1, $svcList[$i])) {
         $num_shown++ 
      }
   }

   my $msg = '';
   if (($num_shown == ($num_msglines-1)) && defined($svcList[$i+1])) { 
      $msg = 'more';
      }
   if ($listoffs > 0) {
      $msg .= ',' if ($msg ne '');
      $msg .= "+$listoffs";
      }
   mLine($num_msglines+$first_msgline-2, AT('botline'), "($msg)")
      if ($msg ne '');

   for ($i=$num_shown+1; $i<$num_msglines; $i++) {
      mLine($i, undef, '~');
      }

   # --- Show total scores
   my $warn_cnt = 0;
   my $crit_cnt = 0;
   my $s1 = sprintf(">Nodes up/dn/unreach: %d/%d/%d",
                    $hState{'UP'},
                    $hState{'DOWN'},
                    $hState{'UNREACHABLE'});
   my $line1 = $s1;

   $crit_cnt += $hState{'DOWN'};
   $warn_cnt += $hState{'UNREACHABLE'};

   $s1 = sprintf(">Svcs ok/warn/crit/other: %d/%d/%d/%d",
                 $sState{'OK'}, $sState{'WARNING'}, $sState{'CRITICAL'},
                 $sState{'UNKNOWN'} + $sState{'PENDING'} + $sState{'HOST DOWN'});

   $warn_cnt += $sState{'WARNING'} + $sState{'PENDING'} + $sState{'UNKNOWN'};
   $crit_cnt += $sState{'CRITICAL'} + $sState{'HOST DOWN'};

   my $attr;
   if ($crit_cnt > 0) {
      $attr = AT('critical');
   } elsif ($warn_cnt > 0) {
      $attr = AT('warning');
   } else {
      $attr = AT('total');
   }

   aat(1, $BOTLINE-1, $attr, $line1);
   aat(1, $BOTLINE, $attr, $s1);

   $last_service_tm = time();
   $display_dirty = 0;
   $num_displays++;

} #showServiceList

}


# -------------------------------------------------------------------------

sub disp_data {

   my ($NSLOG) = @_;

   my @st;
   if (!(@st = stat($NSLOG))) {
      botline("Unable to open status log: $!, '$NSLOG'");
      return -1;
   }
   my $mtime = $st[9];
   if ($mtime > $tm_status_log) {
      #TODO: check for error returns from load_ ..
      load_status_log($NSLOG);

      if ($#statusList < 0) {
         aat(5, 1, "Can't open status log!");
         aat(6, 1, $NSLOG);
         return;
      }

      $hState{'UP'} = 0;
      $hState{'DOWN'} = 0;
      $hState{'UNREACHABLE'} = 0;
      $sState{'OK'} = 0;
      $sState{'CRITICAL'} = 0;
      $sState{'WARNING'} = 0;
      $sState{'UNKNOWN'} = 0;
      $sState{'PENDING'} = 0;
      $sState{'HOST DOWN'} = 0;

      prep_status();
   }

   # --- Ok, clear to go ---


   #filter & sorts service list
   my @svcList = getServiceList();
   dm("Got $#svcList + 1 services..");

   #returns true if anything (new) was shown
   showServiceList(@svcList);

} #disp_data


# =========================================================================
#                     M A I N   F L O W   S U P P O R T
# =========================================================================

{ #localize tables

my @montab = (
   "January", "February", "March", "April", "May", "June",
   "July", "August", "September", "October", "November", "December" );

my @daytab = (
   'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday',
   'Friday', 'Saturday');

sub disp_clock {

   my @l = localtime(time);
   #prepend with spaces to erase former long weekday and month names
   my $s = sprintf('>' . (' ' x 5) . "%s %s %d, %04d, %02d:%02d:%02d",
                   $daytab[$l[6]], $montab[$l[4]], $l[3], $l[5] + 1900, $l[2], $l[1], $l[0]);
   aat(0, 0, AT('dim'), $s);

   if ($upd_pos > 0) {
      my $s2 = '  ' . 'upd ' . tdif(time(), $tm_status_log) . ' ';
      aat($COLS-$upd_pos-length($s2), 1, AT('dim'), $s2);
   }
} #disp_clock


} #localization


# -------------------------------------------------------------------------

my $dirty_options = 1;

sub disp_options {

   return unless $dirty_options;

   my $s = sprintf("%s/%s/%s/%s", 
                   (cfg('showall')?'All':'all'),
                   (cfg('details')?'Det':'det'),
                   (cfg('bell')?'Bell':'bell'),
                   (cfg('reverse')?'Rev':'rev'));
   aat(0, $BOTLINE-1, AT('botline'), substr($s, 0, 40));

   $dirty_options = 0;

} #disp_options


# -------------------------------------------------------------------------
# get_appname() - retrieve application name, only needed at startup
#
sub get_appname {

   my($lf) = @_;

   open(STATUS, $lf) || die('Try again please');   #"never" happens!
   my ($ln) = <STATUS>;
   close(STATUS);

   if ($ln =~ /(NetSaint) (.*) Status File/) {
      $ns_app = 'ns';
   } elsif ($ln =~ /(Nagios) (.*) Status File/) {
      $ns_app = 'nag';
   } else {
      botline("*** Invalid or unrecognized statuslog: $lf");
      return;
   }
   $ns_app_long = "$1 $2";

} #get_appname


# -------------------------------------------------------------------------

sub process_key {

   my ($key) = @_;

   return undef unless ($key);

   if ($key eq 'a') {
      cfg('showall', !cfg('showall'));
      $dirty_options = 1;
   } elsif ($key eq 'd') {
      cfg('details', !cfg('details'));
      $dirty_options = 1;
   } elsif ($key eq 'g') {
      cfg('bell', !cfg('bell'));
      $dirty_options = 1;
   } elsif ($key eq 'r') {
      cfg('reverse', !cfg('reverse'));
      $dirty_options = 1;
   } elsif ($key eq 'C') {
      if (!$opt_colour && !has_colors()) {
         botline("Your terminal can't show colours");
      } else {
         $opt_colour = !$opt_colour;
         drawScreen();
         $dirty_options = 1;
         }
   } elsif ($key eq '^') {
      $listoffs = 0;
   } elsif ($key =~ /^(page| )$/) {
      $listoffs += ($num_msglines-1);
   } elsif ($key eq 'b') {
      if (($listoffs -= ($num_msglines-1)) < 0) {
         $listoffs = 0;
      }
   } elsif ($key eq 'D') {
      cfg('debug', !cfg('debug'));
   } elsif ($key eq 'X') {
      show_debug();
      drawScreen();
   } elsif ($key eq sprintf("%c", ord('L')-64)) {
      erase(); refresh();            #ye olde memory -> /dev/null
      drawScreen();
   } elsif ($key eq 'E') {
      xcmd("$ENV{'VISUAL'} cfg('hostcfg'); echo Enter to reload ..; 
           read; cfg('reloadcmd')");             #debug
   } elsif ($key =~ /[h?]/) {
      help();
      drawScreen();
      $dirty_options = 1;
   } else {
      my $k = 'usercmd_' . $key;
      if (cfg($k)) {
         xcmd(cfg($k));
      } else {
         return undef;
      }
   }
   return 1;

} #process_key


# -------------------------------------------------------------------------

{

my $last_roll = time(); #now, to avoid immediate roll on startup

sub roller {

   my ($last_input) = @_;

   #TODO: don't roll, if we have <=1 page of services...
   if ((cfg('roll') > 0)
       && ((time() - $last_input) >= cfg('roll_wait'))
       && ((time() - $last_roll) >= cfg('roll'))
      ) {
      process_key('page');          
      $display_dirty = 1;
      dm('rolled! ' . sprintf("%d secs since last roll", time()-$last_roll));
      $last_roll = time();
   }

} #roller

} #roller data


# -------------------------------------------------------------------------

#not a complete success, 07-Mar-2004/shj
#sub resize_handler {
#
#   botline(sprintf('window resize to %d,%d', $COLS, $LINES));
#   refresh();
#   return 0;
#
#} #resize_handler
#
#$SIG{'WINCH'} = \&resize_handler;


# =========================================================================
#                M A I N   I N I T I A L I Z A T I O N
# =========================================================================

dm('Greetings and salutations');

$| = 1;           #unbuffered tty i/o
&LoadConfig();

chomp($myself = `hostname`);
$myself =~ s/\..*$//g;

{
   my @s = getpwuid($>);
   $myself = $s[0] . '@' . $myself;
}

#override .config w/cmdline arg if given
my $opt_logfile = defined($ARGV[0])?$ARGV[0]:cfg('nslog');

while (! -f $opt_logfile) {
   print "\nCan't open status.log ($opt_logfile)!\n" .
         "Enter location of status.log: ";
   chomp($opt_logfile = <STDIN>);

   #only set in .config if not commandline supplied and exists
   cfg('nslog', $opt_logfile)
      if (-f $opt_logfile && !defined($ARGV[0]));

} #can't open

&get_appname($opt_logfile);

screen_init();
&drawScreen();

my $cpu = undef;
if ($cpusage) {
   #create this late, so startup costs won't be included
   $cpu = new gothix::cpusage;
   }

#for profiling
my $num_iterations = 0;

#watch out for window resizes
my $last_nlines = $LINES;


# =========================================================================
#                           M A I N   F L O W
# =========================================================================

dm('Going to operational state');

my $last_input = 0;
my $key;
my $done = 0;
while (!$done) {

   #User resized our window? (horizontal resizes are handled elsewhere)
   if ($LINES != $last_nlines) {
      screen_reset();
      drawScreen();
      $last_nlines = $LINES;
   }

   if (cfg('cpumeter') || cfg('debug')) {
      my $s;
      if (defined($cpu)) {
         $s = $cpu->read();
      } else {
         $s = "CPUmeter unavail";
      }
      $s = $s . sprintf(' - #%d', $num_displays) . '    ';
      aat(0, 1, AT('dim'), $s);
   }

   #show curr config before status draw (clreol)
   disp_options();

   #main, live, moving, squirming display!
   disp_data($opt_logfile);

   disp_clock();

   #don't read keys if profiling
   if (defined($ENV{'NSC_ITERATIONS'})) {
      if (++$num_iterations > $ENV{'NSC_ITERATIONS'}) {
         $done = 1;
         last;
      }
      $key = rdKey(0);
   #read once a second to keep clock ticking
   } elsif ($key = rdKey(1)) {
      botline('');                  #clear message on input (?)
      $display_dirty = 1;
      dm(sprintf("got key '%s' %d", $key, ord($key)));
      }

   $last_input = time() if $key;

   # --- Keys handling

   if (!process_key($key)) {
      if ($key && $key eq 'q') {
         botline("Please come back soon");
         $done = 1;
      } elsif ($key && $key eq 'V') {
         xcmd("$ENV{'PAGER'} $opt_logfile");
      } else {
         if (cfg('debug')) {
            dm(sprintf("($num_msglines,$first_msgline,$LINES) (%d/%d)", 
                       AT('normal'),AT('botline')));
            }
         if (!cfg('debug')) {
            botline(sprintf("Press 'h' for help"));
         }
         roller($last_input) if (cfg('roll') > 0);
      }
   }
  
} #while not done

dm('Shutting down operations');

botline('See you!');
screen_down();

&SaveConfig();
print "\n";
exit 0;


# --- End of file ---------------------------------------------------------

# =========================================================================
#                           A L L   D O N E  
# =========================================================================


__END__

